#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <inttypes.h>
#include <assert.h>

#include "fes.h"
#include "idx_LUT.h"
#include "rand.h"


int  report_solution(void *state, uint64_t i) {
  assert(state == NULL);
  printf("%016" PRIx64 "\n", i);
  return 0;
}


////////////////////////

int main(int argc, char** argv) {

  //  srand(time(0));

  int n = 19;
  int n_eqs = 16;
  int random_seed = 0;
  int degree = 2;
  int verbose = 0;

  if (argc > 1) {
    n = atoi(argv[1]);
    n_eqs = atoi(argv[2]);
    degree = atoi(argv[3]);
    verbose = atoi(argv[4]);    
  }
  
  if (argc > 6) random_seed = atoi(argv[5]);

  if (n > 64 || n_eqs > 32) {
    printf("This hack can only handle up to 64 vars and 32 equations\n");
    exit(1);
  }
 
  srand(random_seed);
 // initialize lookup tables
  idx_lut_t* idx_table = init_deginvlex_LUT(n, degree);
  LUT_t LUT = idx_table->LUT;

  // initalize a random system
  const int N = n_monomials(n, degree);
  pck_vector_t *F = calloc(N, sizeof(pck_vector_t)); 
  if (F == NULL) {
    perror("impossible to allocate memory for the coefficients");
    exit(1);
  } 

  for(int i=0; i<N; i++)
    F[i] = myrand() & ((1ll << n_eqs) - 1);

  uint64_t start = rdtsc();

  switch (degree) {
  case 2: exhaustive_ia32_deg_2(LUT, n, F, &report_solution, NULL, verbose); break;
  case 3: exhaustive_ia32_deg_3(LUT, n, F, &report_solution, NULL, verbose); break;
  case 4: exhaustive_ia32_deg_4(LUT, n, F, &report_solution, NULL, verbose); break;
  case 5: exhaustive_ia32_deg_5(LUT, n, F, &report_solution, NULL, verbose); break;
  case 6: exhaustive_ia32_deg_6(LUT, n, F, &report_solution, NULL, verbose); break;
  case 7: exhaustive_ia32_deg_6(LUT, n, F, &report_solution, NULL, verbose); break;
  case 8: exhaustive_ia32_deg_6(LUT, n, F, &report_solution, NULL, verbose); break;
  case 9: exhaustive_ia32_deg_6(LUT, n, F, &report_solution, NULL, verbose); break;
  case 10: exhaustive_ia32_deg_6(LUT, n, F, &report_solution, NULL, verbose); break;
  default:
    printf("enumeration in degree %d is not yet implemented\n", degree);
    exit(0);
  }

  fprintf(stderr, "%.2f cycles/candidate\n", (rdtsc() - start) * 1.0 / (1ll << n));

  free_LUT(idx_table);
  return 0;


 
}
