#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <inttypes.h>
#include <assert.h>

#include "fes.h"
#include "idx_LUT.h"
#include "rand.h"


int  report_solution(void *state, uint64_t i) {
  assert(state == NULL);
  printf("%016" PRIx64 "\n", i);
  return 0;
}


////////////////////////

int main(int argc, char** argv) {

  //  srand(time(0));

  int n = 19;
  int n_eqs = 16;
  int random_seed = 0;
  int degree = 2;

  if (argc > 1) {
    n = atoi(argv[1]);
    n_eqs = atoi(argv[2]);
    degree = atoi(argv[3]);
  }
  
  if (argc > 5) random_seed = atoi(argv[4]);

  if (n > 64 || n_eqs > 32) {
    printf("This hack can only handle up to 64 vars and 32 equations\n");
    exit(1);
  }
 
  srand(random_seed);
 // initialize lookup tables
  idx_lut_t* deginvlex_LUT = init_deginvlex_LUT(n, degree);
  idx_lut_t* lex_LUT = init_lex_LUT(n);


  // initalize a random degree-d system
  const int N = n_monomials(n,degree);
  pck_vector_t *F = calloc(1 << n, sizeof(pck_vector_t));
  for(int i=0; i<N; i++)
    F[i] = myrand() & ((1ll << n_eqs) - 1);

  // scatter it to the full array of all possible monomials 
  pck_vector_t *A = calloc(1 << n, sizeof(pck_vector_t));
  if (F == NULL || A == NULL) {
    printf("Not enough RAM to perform this test. Try a smaller number of variables\n");
    exit(1);
  }
  for(int i=0; i<N; i++)
    A[ idx_convert(deginvlex_LUT, lex_LUT, i) ] = F[i];
  

  uint64_t start = rdtsc();

  fprintf(stderr, "start\n");

  moebius_transform(n, A, &report_solution, NULL);

  fprintf(stderr, "%.2f cycles/candidate\n", (rdtsc() - start) * 1.0 / (1ll << n));

  free_LUT(deginvlex_LUT);
  free_LUT(lex_LUT);
  return 0;
}
